import math
from typing import Callable, List

import numpy as np
from numpy.typing import NDArray


class PenaltyPoint:
    N1 = 3
    N2 = 3
    N3 = 40
    N4 = 10


def get_masks() -> List[Callable]:
    """
    Return the mask function for the given mask pattern_reference.
    """
    pattern_reference_map = [
        lambda i, j: (i + j) % 2 == 0,
        lambda i, j: i % 2 == 0,
        lambda i, j: j % 3 == 0,
        lambda i, j: (i + j) % 3 == 0,
        lambda i, j: (math.floor(i / 2) + math.floor(j / 3)) % 2 == 0,
        lambda i, j: (i * j) % 2 + (i * j) % 3 == 0,
        lambda i, j: ((i * j) % 2 + (i * j) % 3) % 2 == 0,
        lambda i, j: ((i * j) % 3 + (i + j) % 2) % 2 == 0,
    ]
    return pattern_reference_map


def _calculate_adjacent_penalty_inline(array: List[int]) -> int:
    consecutive_run = 1
    score = 0
    for i, num in enumerate(array):
        if i == 0:
            continue

        previous_number = array[i - 1]

        if num == previous_number:
            consecutive_run += 1
            if consecutive_run == 5:
                score += PenaltyPoint.N1
            elif consecutive_run > 5:
                score += 1
        else:
            consecutive_run = 1

    return score


def get_adjacent_modules_penalty(grid: NDArray) -> int:
    result = 0
    for row in grid.tolist():
        result += _calculate_adjacent_penalty_inline(row)

    for col in grid.T.tolist():
        result += _calculate_adjacent_penalty_inline(col)

    return result


def _calculate_finder_penalty(array: NDArray) -> int:
    pattern = [1, 0, 1, 1, 1, 0, 1]
    light_area = [0, 0, 0, 0]

    pattern1 = light_area + pattern
    pattern2 = pattern + light_area

    total_pattern_length = len(pattern1)

    patterns_found = 0

    for i in range(len(array) - total_pattern_length + 1):
        sub_array = array[i : i + total_pattern_length]
        if np.array_equal(sub_array, pattern1):
            patterns_found += 1

        if np.array_equal(sub_array, pattern2):
            patterns_found += 1

    result = patterns_found * PenaltyPoint.N3
    return result


def _calculate_finder_penalty_2(array: NDArray) -> int:
    pattern = [1, 0, 1, 1, 1, 0, 1]
    light_area = [0, 0, 0, 0]

    pattern1 = light_area + pattern
    pattern2 = pattern + light_area

    total_pattern_length = len(pattern1)

    patterns_found = 0

    for i in range(len(array) - total_pattern_length + 1):
        sub_array = array[i : i + total_pattern_length]
        if np.array_equal(sub_array, pattern1):
            patterns_found += 1

        if np.array_equal(sub_array, pattern2):
            patterns_found += 1

    result = patterns_found * PenaltyPoint.N3
    return result


def get_finder_pattern_penalty(grid):
    result = 0
    for row in grid:
        result += _calculate_finder_penalty_2(row)

    for col in grid.T:
        result += _calculate_finder_penalty_2(col)

    return result


def get_same_color_block_penalty(grid: NDArray):
    blocks_count = 0
    grid_size = grid.shape[0]
    for row in range(grid_size - 1):
        for col in range(grid_size - 1):
            if grid[row][col] == grid[row][col + 1] == grid[row + 1][col] == grid[row + 1][col + 1]:
                blocks_count += 1
    result = blocks_count * PenaltyPoint.N2
    return result


def get_proportion_penalty(grid):
    black_proportion = (grid == 1).sum() / grid.size
    deviation = abs((black_proportion * 100) - 50)

    # Every 5% departure from 50%, rating++
    rating = int(deviation / 5)
    result = rating * PenaltyPoint.N4
    return result



import itertools
import re

from qpyr.error_correction import add_ecc_and_interleave, get_num_raw_data_modules
from qpyr.utils import (
    bits_to_bytearray,
    bytearray_to_bits,
    get_segment_character_bits_length,
    get_total_data_capacity_bytes,
)
from qpyr.static import ECC_CODEWORDS_PER_BLOCK, NUM_ERROR_CORRECTION_BLOCKS


def get_best_mode(data: str) -> str:
    numeric_regex: re.Pattern = re.compile(r"[0-9]*")
    alphanumeric_regex: re.Pattern = re.compile(r"[A-Z0-9 $%*+./:-]*")
    is_byte = lambda data: all(ord(char) < 256 for char in data)  # ISO-8859-1 characters

    if numeric_regex.fullmatch(data):
        return "numeric"
    elif alphanumeric_regex.fullmatch(data):
        return "alphanumeric"
    elif is_byte(data):
        return "byte"
    else:
        raise ValueError("Mode not supported")


def get_segment_data(data):
    data_segment = "".join([bin(x)[2:].zfill(8) for x in bytearray(data, "utf-8")])
    return data_segment


def get_segment_mode(mode):
    return {"numeric": "0001", "alphanumeric": "0010", "byte": "0100"}[mode]


def get_segment_character_count(data_segment: str, mode: str, version: int):
    result = ""
    no_of_bytes = len(data_segment) // 8
    if mode == "byte":
        if version <= 9:
            result = bin(no_of_bytes)[2:].zfill(8)
        else:
            result = bin(no_of_bytes)[2:].zfill(16)
    return result


def get_segment_terminator(data_segment: str, mode_segment: str, chr_count_segment: str) -> str:
    total_bits = len(data_segment) + len(mode_segment) + len(chr_count_segment)
    if total_bits % 8 == 4:
        return "0000"
    return ""


def combine_segments(segments: list[str]):
    return "".join(segments)


def get_best_version(data_segment: str, mode: str, ecl: str) -> int:
    total_mode_bits = 4
    bits_required = len(data_segment) + total_mode_bits

    for version in range(1, 41):
        total_capacity_bits = get_total_data_capacity_bytes(ecl, version) * 8
        segment_character_bits_length = get_segment_character_bits_length(mode, version)
        bits_required_total = bits_required + segment_character_bits_length

        if bits_required_total <= total_capacity_bits:
            return version
    raise ValueError("Data too long")


def add_padding(data: str, version: int, ecl: str) -> str:
    """Add padding to data segment.

    Args:
        data (str): data segment

    Returns:
        str: padded data segment
    """
    data_length = len(data)
    bit_padding_required = (8 - (data_length % 8)) % 8
    data = data + "0" * bit_padding_required

    total_num_of_codewords: int = get_num_raw_data_modules(version) // 8
    num_of_blocks = NUM_ERROR_CORRECTION_BLOCKS[ecl][version]
    blockecclen: int = ECC_CODEWORDS_PER_BLOCK[ecl][version]

    total_data_capacity = total_num_of_codewords - (blockecclen * num_of_blocks)
    redundant_bytes_required = ((total_data_capacity * 8) - len(data)) // 8

    padding_bytes = itertools.cycle(["11101100", "00010001"])
    for _ in range(redundant_bytes_required):
        data += next(padding_bytes)
    return data


def encode(data: str, ecl: str):
    """Create a QR code from data.

    Args:
        data (str): data to encode
    """
    mode = get_best_mode(data)
    if mode != "byte":
        raise NotImplementedError("Only byte mode supported.")

    data_segment = get_segment_data(data)
    version = get_best_version(data_segment, mode, ecl)
    mode_segment = get_segment_mode(mode)
    chr_count_segment = get_segment_character_count(data_segment, mode, version)
    terminator_segment = get_segment_terminator(data_segment, mode_segment, chr_count_segment)
    segment = combine_segments([mode_segment, chr_count_segment, data_segment, terminator_segment])
    segment_with_padding = add_padding(segment, version, ecl)
    data_to_encode = bits_to_bytearray(segment_with_padding)
    encoded_data = add_ecc_and_interleave(version=version, ecl=ecl, data=data_to_encode)
    all_bits = bytearray_to_bits(encoded_data)

    return version, all_bits
